/* File yasep/core-js/yasep_disasm.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created 2006-09-04 by whygee@f-cpu.org
2006-11-10 (corrected an edition glitch ???)
2007-04-04 : update for the Dx pseudo-instructions
2007-04-09 : separation of forms and flags
2007-04-17 : another major update
2007-11-05 : rebranding to YASEP, support of the flags IMM16_5LSB and IMM16_8LSB
2009-07-30 : Y and others
2009-08-10 : rebuild !
2009-08-15 : RIR=>IRR, +decode_extended
2009-08-18 : factoring ? trying a new approach where the 5 main forms are separated,
              and 5 functions try to infer the aliases and short versions
2009-08-23 : working again
2011-09-09 : new object model, add_message part of the context
2011-11-06 : Yasep_DISASM=>Y.DISASM, DISASM_context => Y.DISASM_context
2011-11-30 : decode_condition redesigned
2012-03-15 : changed the conditions
2012-10-10 : found an ooold bug in barely tested cases... I need a fuzzer soon !
2013-07-21 : adding FORM_I

This file does the disassembly (decoding) of encoded instructions.
It transforms a number into a string and a size, for interactive debug and testing.

dependency:
 yasep/JScore/txt_int.js,
///////////// (yasep/JSgui/yasep_messages.js+handling [define an empty function if undesired]),
 yasep/JScore/yasep_forms.js/flags/groups/fields
 yasep/JScore/yasep_opcodes.js+children (group_asu.js etc.)
*/

// This is the constructor of the object that is being passed to and from the assembler
Y.DISASM_context=function(add_message) {
  // input
  if (typeof add_message=="function")
    this.add_message = add_message;
  else
    this.add_message = emptyFunc;
  this.YASEP_TYPE=0; // YASEP16/YASEP32 ?
};

// Just a big function with several local sub-functions
Y.DISASM=function(input, t){

  function garbage(a,field){
    if (a != 0)
      t.add_message("yellow", I8N("W")+I8N("dsmgb")+field);
  }

  function decode_short(t,f,g){
// decode the Imm4 forms first

    var i4=t.si4;
    if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
      i4 -= 16;

    if (t.input & Y.FIELD_EXTFORM){
      // aliases :
      if (t.si4==0){
        if (t.name=="SUB"){
          t.name="NEG";
          t.instruction=" "+Y.registers_names[t.snd];
          return "FORM_R"
        }
        if (t.name=="ORN"){
          t.name="NOT";
          t.instruction=" "+Y.registers_names[t.snd];
          return "FORM_R"
        }
      }
      if (f & Y.FORM_iR){
        t.instruction=" "+i4+" "+Y.registers_names[t.snd];
        return "FORM_iR"
      }
      if(f & Y.FORM_Ri){
        t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
        return "FORM_Ri"
      }
    }
    if (f & Y.FORM_i){ // CRIT, HALT

      if ((f & Y.FORM_ALONE)&&(t.si4==0)) { // HALT : optional SI4
        garbage(t.snd,"SND");
        return "FORM_ALONE"
      }

      t.instruction=" "+i4;
      garbage(t.snd,"SND");
      return "FORM_i"
    }

    if (f & Y.FORM_ALONE) {
      garbage(t.si4,"SI4");
      garbage(t.snd,"SND");
      return "FORM_ALONE"
    }

    // some ALONE are already filtered out but not the aliases :
    if (t.input==0) {
      t.name="NOP";
      return "FORM_ALONE"
    }

    // RR :
    t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd];
    return "FORM_RR"
  };

  //////////////////////////// IRR ////////////////////////////
  function decode_long(t,f,g){
    var l=(t.input>>16)& F4;
    var s=16;
    if ((g & Y.IMM20) && (t.YASEP_TYPE!=16)) {
      s=20;
      l |= t.snd<<16;
    }

    if (!(g & Y.IgnoreImmSign) && (l >= (1<<(s-1))))
      l -= (1<<s);

    if (g & Y.IGNORE_IMM16){ // does it ignore the Imm16 field ?

      if (f&Y.FORM_I) { // HALT
        garbage(t.si4,"SI4");
        garbage(t.snd,"SND");
        t.instruction=" "+l;
        return "FORM_I"
      }

      t.form="FORM_ALONE";
      garbage(l,"IMM16");
      garbage(t.snd,"SND");
      garbage(t.si4,"SI4");
    }

    if (f&Y.FORM_RI) { // PUT
      garbage(t.si4,"SI4");
      t.instruction=" "+Y.registers_names[t.snd]+" "+l;
      return "FORM_RI"
    }   

    if (f&Y.FORM_IR) {
      if (g&Y.ALIAS_IRR){
        if (t.snd==t.si4) { // condensed writing
          t.instruction=" "+l+" "+Y.registers_names[t.snd]; // either si4 or snd, they are equal...
          return "FORM_IR"
        }
      }
      else {
        if (g & Y.SWAP_IR){ // MOV, GET
          if (s!=20)
            garbage(t.snd,"SND");
          t.instruction=" "+l+" "+Y.registers_names[t.si4];
        }
        else { // CMPU, CMPS
          garbage(t.si4,"SI4");
          t.instruction=" "+l+" "+Y.registers_names[t.snd];
        }
        return "FORM_IR"
      }
    }
    // default
    t.instruction=" "+l+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.si4];
    return "FORM_IRR"
  }

  function decode_extended(t,f,g){
  // decode the Imm4 forms first
    if (t.input & Y.FIELD_EXTIMM4) {
      var i4=t.si4;
      if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
        i4 -= 16;

      // aliases :
      if ((t.si4==0)&&(t.snd==t.dst3)){
        if (t.name=="SUB"){
          t.name="NEG";
          t.instruction=" "+Y.registers_names[t.snd];
          if (t.snd == t.dst3)
            return "FORM_R";
          t.instruction=" "+Y.registers_names[t.dst3];
          return "FORM_RR"
        }
        if (t.name=="ORN"){
          t.name="NOT";
          t.instruction=" "+Y.registers_names[t.snd];
          if (t.snd == t.dst3)
            return "FORM_R";
          t.instruction=" "+Y.registers_names[t.dst3];
         return "FORM_RR"
        }
      }
      if ((f & Y.FORM_iR)&&(t.snd == t.dst3)) {
        t.instruction=" "+i4+" "+Y.registers_names[t.snd];
        return "FORM_iR"
      }
      if(f & Y.FORM_Ri){ // PUT
        t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
        return "FORM_Ri"
      }
      // default
      if (f & Y.FORM_iRR) {

        if (t.snd == t.dst3) {
          t.instruction=" "+i4+" "+Y.registers_names[t.snd];
          return "FORM_iR"
        }

        t.instruction=" "+i4+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
        return "FORM_iRR"
      }
    }
    else {
      if ((f & Y.FORM_RR)&&(t.snd == t.dst3)) {
        t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.dst3];
        return "FORM_RR"
      }
      if (f & Y.FORM_RRR) {
        t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
        return "FORM_RRR"
      }
    }
    t.add_message("red", I8N("R")+I8N("dsmuk"));
  }

  ConditionPC=["", // "ALWAYS"
  " NEVER",
  " BIT1 0",
  " BIT0 0",
  " CARRY",
  " NO_CARRY",
  " EQ",
  " NEQ"];

  function decode_condition(t){
    var r, cond=(t.input & Y.FIELD_EXTCOND)>>Y.OFFSET_EXTCOND;

    // PC ?
    if (t.src4==0)
      return ConditionPC[cond];

    // else other cases
    r=Y.registers_names[t.src4];
    switch(cond) {
      case 0: return " ZERO "  +r;
      case 1: return " NZ "  +r;
      case 2: return " BIT1 "+t.src4;
      case 3: return " BIT0 "+t.src4;
      case 4: return " LSB1 "+r;
      case 5: return " LSB0 "+r;
      case 6: return " MSB1 "+r;
      case 7: return " MSB0 "+r;
    }
  }


////////////////////////////////////////////////////
// Start of the decoding function

// t.decode_instruction=function(input) {
  t.form=null; // indicates the error if still null after return
  t.prefix=input & (Y.FIELD_IMMREG|Y.FIELD_EXTFORM);
  t.opcode=input &  Y.FIELD_OPCODE;
  t.input=input;
  t.instruction="";
  t.size_bytes=0;
  t.snd=(input & Y.FIELD_SND)>>Y.OFFSET_SND;
  t.si4=(input & Y.FIELD_SI4)>>Y.OFFSET_SI4;
  t.name=null;

  if (typeof input != "number") {
    t.add_message("red", I8N("R")+I8N("dsmif"));
    return
  }

  // opcode lookup
  t.O=Y.opcode_table[t.opcode];

  if (typeof t.O != "object") {
    if (input & Y.FIELD_IMMREG) {
      // output a DW
      t.add_message("red", I8N("R")+I8N("dsmuo")+int2hex(t.opcode)+I8N("dsmuol"));
      t.instruction = "DW "+ int2hex(input,7,7);
      t.size_byte=4;
    }
    else {
      // output a DH
      t.add_message("red", I8N("R")+I8N("dsmuo")+int2hex(t.opcode)+I8N("dsmuos"));
      garbage(input & Y.FIELD_IMM16,"the MSB");
      t.instruction = "DH "+ int2hex(input & F4);
      t.size_byte=2;
    }
    return
  }

  t.name=t.O.name;
  var f=t.O.forms;
  var g=t.O.flags;

  // the size is straightforward to find :
  if (input & Y.FIELD_IMMREG) {
    // here we process a long instruction 
    t.size_bytes=4;

    // extended forms :
    if (input & Y.FIELD_EXTFORM) {
      t.dst3=(input>>Y.OFFSET_EXTDST3)&15;
      t.src4=(input>>Y.OFFSET_EXTSRC4)&15;
      t.form=decode_extended(t,f,g);
      t.instruction+=decode_condition(t);
    }
    else { // Long Immediate instructions : FORM_RI, FORM_IR, FORM_IRR
      // test the garbage :
      if((input & 0xFF000000)&&(g&Y.IMM16_8LSB)
      || (input & 0xFFE00000)&&(g&Y.IMM16_5LSB))
        t.add_message("yellow", I8N("W")+I8N("dsmgb")+"IMM16");
      t.form=decode_long(t,f,g);
    }
  }
  else { // Short forms
    t.size_bytes=2;
    t.form=decode_short(t,f,g);
  }

  // output
  if (t.form!=null){
    t.instruction=t.name+t.instruction;
    t.add_message("blue"," OK: opcode='"+Y.decorate_opcode(t.opcode)+" instruction=<b>"+t.instruction+"</b> ("+t.form+")");
  }
};
