/* yasep/core-js/yasep_profiles.js
  This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
dim. févr. 19 19:23:01 CET 2012 : moved core from tool/config.js
jeu. sept. 13 00:50:45 CEST 2012 : uses core-js/txt_int.js
sam. oct.  6 19:24:17 CEST 2012 : renamed from yasep_config.js, removed YASEP (that was confusing)
mer. oct. 10 02:28:44 CEST 2012 : added .valid for easier decoding, but not yet integrated in VHDL/config
lun. juil. 29 06:36:33 CEST 2013 : patching a bit...

This file contains various "core profiles" of varying implementations
and some supporting functions.

Note:
usually this is where profiles are created.
To reduce the code source size, there is code
that transforms the mask into a full object.
The mask is created by going through #!VHDL/configure
and exporting the VHDL of a given profile,
the mask is in the comments.
Well, the configure tool is quite buggy (it is being
rebuilt) and the mask must be changed when instructions
are added or modified so I "bootstrap" the profiles
by restoring the original instruction list object.
 */

/*
blacklist(PC_length);
blacklist(datapath);
blacklist(multiplier);
blacklist(pipelined);
blacklist(Enable_BIT0);
blacklist();
*/

// turns a sparse array in a solid, fixed-sized array and string
function collapse(a) {
  for (var i=0; i<64; i++)
    if (typeof a[i]!="number")
      a[i]=0;
  return a.join("");
}

// Takes a string of 64 ones and zeroes (bit #0 on the left, at index 0,
//  range "to" in VHDL) and makes a 16-char hexa string
// Problem : integers are limited to about 50 bits, not 64, so it is processed
// in chunks of 4 bits
function encodeInstructionMask(b){
  var o="",i,j=0,c;
  if (b.length!=64)
    return alert("decodeInstructionMask : wrong hex string size ("+h+")");

  while (j<64){
    c=0;
    for (i=0;i<4;i++)
      c+=c+(b[j++]&1); // shift left
    o+=hex_array[c];
  }
  return o;
}

// Take a hexa string, decompose in binary and rebuild an object from it
// with a collection of opcode names
function decodeInstructionMask(h){
  var m=h.mask, o={},c,i=0,j=0, v=[];
  if (m.length!=16)
    return alert("decodeInstructionMask : wrong hex string size ("+m+")");

  while (i<256){
    if(!(i&12)) // 2 LSB=0 => fetch a new character from h
      c=int_array[m[j++]];

    // decode
    v[i]=0;
    if ((c&8) && Y.opcode_table[i])
      o[Y.opcode_table[i].name]
        =v[i]
        =1;

    c+=c; // shift left
    i+=4; // 256/4=64 opcodes
  }
  h.valid=v;
  h.valid_instructions=o;
}

profiles={
// This object gets scanned, so don't put funky stuff in here.

  "YASEP32":{
    comment: "wide datapath, full-featured version",
    datapath:32,
    PC_length:32,
    multiplier:16,
    CondR1:1,
    mask:"FEFEFEFEBEAEACAD"
/*  valid_instructions: {
////// GROUP_CTL :
      "NOP":1,
      "CRIT":1,
      "GET":1,
      "PUT":1,
      "IN":1,
      "OUT":1,
      "CALL":1,
      "CALL2":1,

////// GROUP_SMT :
      "HALT":1,
      "IPC":1,
      "IPE":1,
      "IPR":1,

////// GROUP_IE :
      "MOV":1,
      "IB":1,
      "EZB":1,
      "ESB":1,
      "MOVH":1,
      "IH":1,
      "EZH":1,
      "ESH":1,

////// GROUP_SHL :
      "SHR":1,
      "SAR":1,
      "SHL":1,
      "ROR":1,
      "ROL":1,

////// GROUP_ROP2 :
      "AND":1,
      "ANDN":1,
      "NAND":1,
      "OR":1,
      "ORN":1,
      "NOR":1,
      "XOR":1,
      "XORN":1,

////// GROUP_ASU :
      "ADD":1,
      "SUB":1,
      "CMPU":1,
      "CMPS":1,
      "UMIN":1,
      "UMAX":1,
      "SMIN":1,
      "SMAX":1,

////// GROUP_MUL :
      "MUL8L":1,
      "MUL8H":1,
      "MULI":1,
      "LUT8":1,
      "MUL16L":1,
      "MUL16H":1,

////// GROUP_RSV :
      "INV":1
    }
*/
  },

  "YASEP16":{
    comment: "narrow datapath version",
    datapath:16,
    PC_length:16,
    multiplier:8,
    CondR1:1,
    mask:"FEFEFEFE9C8C8C8D"
/*
    valid_instructions: {
////// GROUP_CTL :
      "NOP":1,
      "CRIT":1,
      "GET":1,
      "PUT":1,
      "IN":1,
      "OUT":1,
      "CALL":1,
      "CALL2":1,

////// GROUP_SMT :
      "HALT":1,
      "IPC":1,
      "IPE":1,
      "IPR":1,

////// GROUP_IE :
      "MOV":1,
      "IB":1,
      "EZB":1,
      "ESB":1,

////// GROUP_SHL :
      "SHR":1,
      "SAR":1,
      "SHL":1,
      "ROR":1,
      "ROL":1,

////// GROUP_ROP2 :
      "AND":1,
      "ANDN":1,
      "NAND":1,
      "OR":1,
      "ORN":1,
      "NOR":1,
      "XOR":1,
      "XORN":1,

////// GROUP_ASU :
      "ADD":1,
      "SUB":1,
      "CMPU":1,
      "CMPS":1,
      "UMIN":1,
      "UMAX":1,
      "SMIN":1,
      "SMAX":1,

////// GROUP_MUL :
      "MUL8L":1,
      "MUL8H":1,
      "MULI":1,
      "LUT8":1,

////// GROUP_RSV :
      "INV":1
    }
*/
  },

  "microYASEP":{
    comment: "narrow, minimal core with limited data memory access",
    datapath:16,
    multiplier:0,
    PC_length:16,
    CondR1:1,
    mask:"AC0C8C8C08088809"

/*    valid_instructions: {
////// GROUP_CTL :
      "NOP":1,
      "GET":1,
      "PUT":1,
      "CALL":1,

////// GROUP_IE :
      "MOV":1,

////// GROUP_ROP2 :
      "AND":1,
      "ANDN":1,
      "NAND":1,
      "OR":1,
      "ORN":1,
      "NOR":1,
      "XOR":1,
      "XORN":1,

////// GROUP_ASU :
      "ADD":1,
      "SUB":1,
      "CMPU":1,
      "CMPS":1,

////// GROUP_RSV :
      "INV":1
    }
*/
  },

  "empty":{
    comment: "empty profile for the 'auto' assembly option that fills the instructions list",
    datapath:16,
    multiplier:0,
    PC_length:8,
    CondR1:0,
    mask:"0000000000000000"
//    valid_instructions: {/*nothing*/}
  }
};

for (var i in profiles){
  if (profiles[i].mask) // !!!!
    decodeInstructionMask(profiles[i]);
}
